import numpy as np


class NaiveCentroid:
  """A cluster centroid of the Naive clustering algorithm."""

  def __init__(self, embedding):
    """Create a new centroid."""
    self.embedding = embedding
    self.count = 1

  def merge(self, embedding):
    """Merge a new embedding into the centroid."""
    self.embedding = (self.embedding * self.count + embedding) / (
        self.count + 1)
    self.count += 1

  def cosine(self, embedding):
    """Compute cosine similarity to a new embedding."""
    return np.dot(self.embedding, embedding) / (
        np.linalg.norm(self.embedding) * np.linalg.norm(embedding))


class NaiveClusterer:
  """Naive clustering class."""

  def __init__(self, threshold, adaptation_threshold=None):
    """Initialized the clusterer.

    Note that since this is online clustering, fit_predict and predict
    are the same.

    Args:
      threshold: if cosine similarity is larger than this threshold, the
        embedding will be considered to belong to the cluster
      adaptation_threshold: if cosine similarity is larger than
        adaptation_threshold, the embedding will be merged to the cluster.
        If None, we use threshold as adaptation_threshold

    Raises:
      ValueError: if adaptation_threshold is smaller than threshold
    """
    self.threshold = threshold
    if adaptation_threshold is None:
      self.adaptation_threshold = threshold
    elif adaptation_threshold < threshold:
      raise ValueError("adaptation_threshold cannot be smaller than threshold")
    else:
      self.adaptation_threshold = adaptation_threshold
    self.centroids = []

  def reset(self):
    """Reset the clusterer."""
    self.centroids = []

  def predict_next(self, embedding):
    """Given a new embedding, output its label.

    This is used for online clustering.

    Args:
      embedding: numpy array of shape (n_features,)

    Returns:
      label: an integer cluster label
    """
    # Handle first embedding case.
    if len(self.centroids) == 0:
      self.centroids.append(NaiveCentroid(embedding))
      return 0

    # Compute all similarities.
    similarities = np.array(
        [centroid.cosine(embedding) for centroid in self.centroids])

    # New cluster.
    if similarities.max() < self.threshold:
      self.centroids.append(NaiveCentroid(embedding))
      return len(self.centroids) - 1

    # Existing cluster.
    label = similarities.argmax()
    if similarities[label] > self.adaptation_threshold:
      self.centroids[label].merge(embedding)
    return label

  def predict(self, embeddings):
    """Given many embeddings, return all cluster labels.

    This is for simulating offline clustering behavior.

    Args:
      embeddings: numpy array of shape (n_samples, n_features)

    Returns:
      labels: numpy array of shape (n_samples,)
    """
    return np.array([self.predict_next(embedding) for embedding in embeddings])

  def fit_predict(self, embeddings):
    """Same as predict(), since this is an online clusterer."""
    return self.predict(embeddings)



"""__init__ file."""

from . import autotune
from . import configs
from . import constraint
from . import fallback_clusterer
from . import laplacian
from . import naive_clusterer
from . import refinement
from . import spectral_clusterer
from . import utils

AutoTune = autotune.AutoTune
AutoTuneProxy = autotune.AutoTuneProxy

ConstraintOptions = constraint.ConstraintOptions
ConstraintName = constraint.ConstraintName
ConstraintMatrix = constraint.ConstraintMatrix
IntegrationType = constraint.IntegrationType

FallbackOptions = fallback_clusterer.FallbackOptions
SingleClusterCondition = fallback_clusterer.SingleClusterCondition
FallbackClustererType = fallback_clusterer.FallbackClustererType

LaplacianType = laplacian.LaplacianType

NaiveClusterer = naive_clusterer.NaiveClusterer

RefinementName = refinement.RefinementName
RefinementOptions = refinement.RefinementOptions
ThresholdType = refinement.ThresholdType
SymmetrizeType = refinement.SymmetrizeType

SpectralClusterer = spectral_clusterer.SpectralClusterer

EigenGapType = utils.EigenGapType

ICASSP2018_REFINEMENT_SEQUENCE = configs.ICASSP2018_REFINEMENT_SEQUENCE
TURNTODIARIZE_REFINEMENT_SEQUENCE = configs.TURNTODIARIZE_REFINEMENT_SEQUENCE



"""Implement custom kmeans.

It supports any distance measure defined in scipy.spatial.distance.
"""

import numpy as np
from scipy.spatial import distance
from sklearn.cluster import KMeans


def run_kmeans(spectral_embeddings, n_clusters, custom_dist, max_iter):
  """Run CustomKMeans with a custom distance measure support.

  Perform a custom kmeans clustering with any distance measure defined
  in scipy.spatial.distance.

  Args:
    spectral_embeddings: input spectracl embedding observations
    n_clusters: the number of clusters to form
    custom_dist: str or callable. custom distance measure for k-means. if a
      string, "cosine", "euclidean", "mahalanobis", or any other distance
      functions defined in scipy.spatial.distance can be used
    max_iter: the maximum number of iterations for the custom k-means

  Returns:
    labels: predicted clustering labels of all samples
  """
  if not custom_dist:  # Scikit-learn KMeans
    kmeans_clusterer = KMeans(
        n_clusters=n_clusters, init="k-means++", max_iter=300, random_state=0)
  else:
    # Initialization using the k-means++ method in Scikit-learn
    kmeans_clusterer = KMeans(
        n_clusters=n_clusters, init="k-means++", max_iter=1, random_state=0)
    kmeans_clusterer.fit(spectral_embeddings)
    centroids = kmeans_clusterer.cluster_centers_
    # Run the cusotom K-means
    kmeans_clusterer = CustomKMeans(
        n_clusters=n_clusters,
        centroids=centroids,
        max_iter=max_iter,
        custom_dist=custom_dist)

  labels = kmeans_clusterer.predict(spectral_embeddings)
  return labels


class CustomKMeans:
  """Class CustomKMeans performs KMeans clustering."""

  def __init__(self,
               n_clusters=None,
               centroids=None,
               max_iter=10,
               tol=.001,
               custom_dist="cosine"):
    """Constructor of the clusterer.

    A custom kmeans clusterer with any distance measure from
    scipy.spatial.distance can be used.

    Args:
      n_clusters: the number of clusters to form
      centroids: the cluster centroids. if given, initial centroids are set as
        the input samples. if not, centroids are randomly initialized
      max_iter: maximum number of iterations of the k-means algorithm to run
      tol: the relative increment in the results before declaring convergence
      custom_dist: str or callable. custom distance measure to use. if a string,
        "cosine", "euclidean", "mahalanobis", or any other distance functions
        defined in scipy.spatial.distance can be used
    """
    self.n_clusters = n_clusters
    self.centroids = centroids
    self.max_iter = max_iter
    self.tol = tol
    self.custom_dist = custom_dist

  def _init_centroids(self, embeddings):
    """Compute the initial centroids."""

    n_samples = embeddings.shape[0]
    idx = np.random.choice(
        np.arange(n_samples), size=self.n_clusters, replace=False)
    self.centroids = embeddings[idx, :]

  def predict(self, embeddings):
    """Performs the clustering.

    Args:
      embeddings: the input observations to cluster

    Returns:
      labels: predicted clustering labels of all samples

    Raises:
      ValueError: if input observations have wrong shape
    """
    n_samples, n_features = embeddings.shape
    if self.max_iter <= 0:
      raise ValueError("Number of iterations should be a positive number,"
                       " got %d instead" % self.max_iter)
    if n_samples < self.n_clusters:
      raise ValueError("n_samples=%d should be >= n_clusters=%d" %
                       (n_samples, self.n_clusters))
    if self.centroids is None:
      self._init_centroids(embeddings)
    else:
      n_centroids, c_n_features = self.centroids.shape
      if n_centroids != self.n_clusters:
        raise ValueError("The shape of the initial centroids (%s)"
                         "does not match the number of clusters %d" %
                         (str(self.centroids.shape), self.n_clusters))
      if n_features != c_n_features:
        raise ValueError(
            "The number of features of the initial centroids %d"
            "does not match the number of features of the data %d." %
            (c_n_features, n_features))

    sample_ids = np.arange(n_samples)
    prev_mean_dist = 0
    for iter_idx in range(self.max_iter + 1):
      # Compute distances to all centroids and assign each sample a label
      # corresponding to the nearest centroid
      dist_to_all_centroids = distance.cdist(
          embeddings, self.centroids, metric=self.custom_dist)
      labels = dist_to_all_centroids.argmin(axis=1)
      distances = dist_to_all_centroids[sample_ids, labels]
      mean_distance = np.mean(distances)
      # If the difference between current mean_distance and previous
      # mean_distance is very small or the max iteration number is reached,
      # the clustering stops
      if mean_distance <= prev_mean_dist and mean_distance >= (
          1 - self.tol) * prev_mean_dist or iter_idx == self.max_iter:
        break
      prev_mean_dist = mean_distance
      # Update centroids
      for each_centroid_idx in range(n_centroids):
        each_centroid_samples = np.where(labels == each_centroid_idx)[0]
        if each_centroid_samples.any():
          self.centroids[each_centroid_idx] = np.mean(
              embeddings[each_centroid_samples], axis=0)
    return labels



"""Affinity matrix refinemnet operations."""

import abc
import enum
import numpy as np
from scipy.ndimage import gaussian_filter


class RefinementName(enum.Enum):
  """The names of the refinement operations."""
  CropDiagonal = enum.auto()
  GaussianBlur = enum.auto()
  RowWiseThreshold = enum.auto()
  Symmetrize = enum.auto()
  Diffuse = enum.auto()
  RowWiseNormalize = enum.auto()


class ThresholdType(enum.Enum):
  """Different types of thresholding."""
  # We clear values that are smaller than row_max*p_percentile
  RowMax = enum.auto()

  # We clear (p_percentile*100)% smallest values of the entire row
  Percentile = enum.auto()


class SymmetrizeType(enum.Enum):
  """Different types of symmetrization operation."""
  # We use max(A, A^T)
  Max = enum.auto()

  # We use 1/2(A + A^T)
  Average = enum.auto()


class RefinementOptions:
  """Refinement options for the affinity matrix."""

  def __init__(self,
               gaussian_blur_sigma=1,
               p_percentile=0.95,
               thresholding_soft_multiplier=0.01,
               thresholding_type=ThresholdType.RowMax,
               thresholding_with_binarization=False,
               thresholding_preserve_diagonal=False,
               symmetrize_type=SymmetrizeType.Max,
               refinement_sequence=None):
    """Initialization of the refinement arguments.

    Args:
      gaussian_blur_sigma: sigma value of the Gaussian blur operation
      p_percentile: the p-percentile for the row wise thresholding
      thresholding_soft_multiplier: the multiplier for soft threhsold, if this
        value is 0, then it's a hard thresholding
      thresholding_type: the type of thresholding operation
      thresholding_with_binarization: if true, we set values larger than the
        threshold to 1
      thresholding_preserve_diagonal: if true, in the row wise thresholding
        operation, we firstly set diagonals of the affinity matrix to 0, and set
        the diagonals back to 1 in the end
      symmetrize_type: a SymmetrizeType
      refinement_sequence: a list of RefinementName for the sequence of
        refinement operations to apply on the affinity matrix. If None, we will
        not refine
    """
    self.gaussian_blur_sigma = gaussian_blur_sigma
    self.p_percentile = p_percentile
    self.thresholding_soft_multiplier = thresholding_soft_multiplier
    self.thresholding_type = thresholding_type
    self.thresholding_with_binarization = thresholding_with_binarization
    self.thresholding_preserve_diagonal = thresholding_preserve_diagonal
    self.symmetrize_type = symmetrize_type
    if refinement_sequence is None:
      self.refinement_sequence = []
    elif not isinstance(refinement_sequence, list):
      raise TypeError("refinement_sequence must be a list of RefinementName")
    else:
      self.refinement_sequence = refinement_sequence

  def get_refinement_operator(self, name):
    """Get the refinement operator for the affinity matrix.

    Args:
      name: a RefinementName

    Returns:
      object of the operator

    Raises:
      TypeError: if name is not a RefinementName
      ValueError: if name is an unknown refinement operation
    """
    if not isinstance(name, RefinementName):
      raise TypeError("name must be a RefinementName")
    elif name == RefinementName.CropDiagonal:
      return CropDiagonal()
    elif name == RefinementName.GaussianBlur:
      return GaussianBlur(self.gaussian_blur_sigma)
    elif name == RefinementName.RowWiseThreshold:
      return RowWiseThreshold(self.p_percentile,
                              self.thresholding_soft_multiplier,
                              self.thresholding_type,
                              self.thresholding_with_binarization,
                              self.thresholding_preserve_diagonal)
    elif name == RefinementName.Symmetrize:
      return Symmetrize(self.symmetrize_type)
    elif name == RefinementName.Diffuse:
      return Diffuse()
    elif name == RefinementName.RowWiseNormalize:
      return RowWiseNormalize()
    else:
      raise ValueError("Unknown refinement operation: {}".format(name))


class AffinityRefinementOperation(metaclass=abc.ABCMeta):
  """Refinement of the affinity matrix."""

  def check_input(self, affinity):
    """Check the input to the refine() method.

    Args:
      affinity: the input affinity matrix.

    Raises:
      TypeError: if affinity has wrong type
      ValueError: if affinity has wrong shape, etc.
    """
    if not isinstance(affinity, np.ndarray):
      raise TypeError("affinity must be a numpy array")
    shape = affinity.shape
    if len(shape) != 2:
      raise ValueError("affinity must be 2-dimensional")
    if shape[0] != shape[1]:
      raise ValueError("affinity must be a square matrix")

  @abc.abstractmethod
  def refine(self, affinity):
    """An abstract method to perform the refinement operation.

    Args:
        affinity: the affinity matrix, of size (n_samples, n_samples)

    Returns:
        a matrix of the same size as affinity
    """
    pass


class CropDiagonal(AffinityRefinementOperation):
  """Crop the diagonal.

  Replace diagonal element by the max non-diagonal value of row.
  After this operation, the matrix has similar properties to a standard
  Laplacian matrix. This also helps to avoid the bias during Gaussian blur and
  normalization.
  """

  def refine(self, affinity):
    self.check_input(affinity)
    refined_affinity = np.copy(affinity)
    np.fill_diagonal(refined_affinity, 0.0)
    di = np.diag_indices(refined_affinity.shape[0])
    refined_affinity[di] = refined_affinity.max(axis=1)
    return refined_affinity


class GaussianBlur(AffinityRefinementOperation):
  """Apply Gaussian blur."""

  def __init__(self, sigma=1):
    self.sigma = sigma

  def refine(self, affinity):
    self.check_input(affinity)
    return gaussian_filter(affinity, sigma=self.sigma)


class RowWiseThreshold(AffinityRefinementOperation):
  """Apply row wise thresholding."""

  def __init__(self,
               p_percentile=0.95,
               thresholding_soft_multiplier=0.01,
               thresholding_type=ThresholdType.RowMax,
               thresholding_with_binarization=False,
               thresholding_preserve_diagonal=False):
    self.p_percentile = p_percentile
    self.multiplier = thresholding_soft_multiplier
    if not isinstance(thresholding_type, ThresholdType):
      raise TypeError("thresholding_type must be a ThresholdType")
    self.thresholding_type = thresholding_type
    self.thresholding_with_binarization = thresholding_with_binarization
    self.thresholding_preserve_diagonal = thresholding_preserve_diagonal

  def refine(self, affinity):
    self.check_input(affinity)
    refined_affinity = np.copy(affinity)
    if self.thresholding_preserve_diagonal:
      np.fill_diagonal(refined_affinity, 0.0)
    if self.thresholding_type == ThresholdType.RowMax:
      # Row_max based thresholding
      row_max = refined_affinity.max(axis=1)
      row_max = np.expand_dims(row_max, axis=1)
      is_smaller = refined_affinity < (row_max * self.p_percentile)
    elif self.thresholding_type == ThresholdType.Percentile:
      # Percentile based thresholding
      row_percentile = np.percentile(
          refined_affinity, self.p_percentile * 100, axis=1)
      row_percentile = np.expand_dims(row_percentile, axis=1)
      is_smaller = refined_affinity < row_percentile
    else:
      raise ValueError("Unsupported thresholding_type")
    if self.thresholding_with_binarization:
      # For values larger than the threshold, we binarize them to 1
      refined_affinity = (np.ones_like(
          (refined_affinity)) * np.invert(is_smaller)) + (
              refined_affinity * self.multiplier * is_smaller)
    else:
      refined_affinity = (refined_affinity * np.invert(is_smaller)) + (
          refined_affinity * self.multiplier * is_smaller)
    if self.thresholding_preserve_diagonal:
      np.fill_diagonal(refined_affinity, 1.0)
    return refined_affinity


class Symmetrize(AffinityRefinementOperation):
  """The Symmetrization operation."""

  def __init__(self, symmetrize_type=SymmetrizeType.Max):
    if not isinstance(symmetrize_type, SymmetrizeType):
      raise TypeError("symmetrize_type must be a SymmetrizeType")
    self.symmetrize_type = symmetrize_type

  def refine(self, affinity):
    self.check_input(affinity)
    if self.symmetrize_type == SymmetrizeType.Max:
      return np.maximum(affinity, np.transpose(affinity))
    elif self.symmetrize_type == SymmetrizeType.Average:
      return 0.5 * (affinity + np.transpose(affinity))
    else:
      raise ValueError("Unsupported symmetrize_type.")


class Diffuse(AffinityRefinementOperation):
  """The diffusion operation."""

  def refine(self, affinity):
    self.check_input(affinity)
    return np.matmul(affinity, np.transpose(affinity))


class RowWiseNormalize(AffinityRefinementOperation):
  """The row wise max normalization operation."""

  def refine(self, affinity):
    self.check_input(affinity)
    refined_affinity = np.copy(affinity)
    row_max = refined_affinity.max(axis=1)
    refined_affinity /= np.expand_dims(row_max, axis=1)
    return refined_affinity



"""Setup script for the package."""

import setuptools

VERSION = "0.2.16"

with open("README.md", "r") as file_object:
    LONG_DESCRIPTION = file_object.read()

setuptools.setup(
    name="spectralcluster",
    version=VERSION,
    author="Quan Wang",
    author_email="quanw@google.com",
    description="Spectral Clustering",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/wq2012/SpectralCluster",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)



"""In some cases, we use a fallback clusterer instead of spectral.

Spectral clustering exploits the global structure of the data. But there are
cases where spectral clustering does not work as well as some other simpler
clustering methods, such as when the number of embeddings is too small.
"""

import enum
import numpy as np
from sklearn.cluster import AgglomerativeClustering
from sklearn.mixture import GaussianMixture
from spectralcluster import naive_clusterer


class SingleClusterCondition(enum.Enum):
  """Which condition do we use for deciding single-vs-multi cluster(s)."""

  # Fit affinity values with GMM with 1-vs-2 component(s), and use
  # Bayesian Information Criterion (BIC) to decide whether there are
  # at least two clusters.
  # Note that this approach does not require additional parameters.
  AffinityGmmBic = enum.auto()

  # If all affinities are larger than threshold, there is only a single cluster.
  AllAffinity = enum.auto()

  # If all neighboring affinities are larger than threshold, there is only
  # a single cluster.
  NeighborAffinity = enum.auto()

  # If the standard deviation of all affinities is smaller than threshold,
  # there is only a single cluster.
  AffinityStd = enum.auto()

  # Use fallback clusterer to make the decision. If fallback clusterer
  # finds multiple clusters, continue with spectral clusterer.
  FallbackClusterer = enum.auto()


class FallbackClustererType(enum.Enum):
  """Which fallback clusterer to use."""

  # AgglomerativeClustering from scikit-learn.
  Agglomerative = enum.auto()

  # Naive clustering, as described in the paper "Speaker diarization with LSTM".
  Naive = enum.auto()


class FallbackOptions:
  """Options for fallback options."""

  def __init__(self,
               spectral_min_embeddings=1,
               single_cluster_condition=SingleClusterCondition.AffinityGmmBic,
               single_cluster_affinity_threshold=0.75,
               single_cluster_affinity_diagonal_offset=1,
               fallback_clusterer_type=FallbackClustererType.Naive,
               agglomerative_threshold=0.5,
               naive_threshold=0.5,
               naive_adaptation_threshold=None):
    """Initialization of the fallback options.

    Args:
      spectral_min_embeddings: we only run spectral clusterer if we have at
        least these many embeddings; otherwise we run fallback clusterer
      single_cluster_condition: how do we decide single-vs-multi cluster(s)
      single_cluster_affinity_threshold: affinity threshold to decide
        whether there is only a single cluster
      single_cluster_affinity_diagonal_offset: when using AffinityGmmBic
        to make single-vs-multi cluster(s) decisions, we only fit the GMM to
        the upper triangular matrix because the diagonal and near-diagonal
        values might be very big. By default, we use a value of 1 to only
        exclude diagonal values. But if embeddings are extracted from
        overlapping sliding windows, this value could be larger than 1
      fallback_clusterer_type: which fallback clusterer to use
      agglomerative_threshold: threshold of agglomerative clustering
      naive_threshold: threshold for naive clusterer
      naive_adaptation_threshold: adaptation_threshold for naive clusterer
    """
    if not isinstance(spectral_min_embeddings, int):
      raise TypeError("spectral_min_embeddings must be an integer")
    if not isinstance(single_cluster_affinity_diagonal_offset, int):
      raise TypeError(
          "single_cluster_affinity_diagonal_offset must be an integer")
    if single_cluster_affinity_diagonal_offset < 0:
      raise ValueError(
          "single_cluster_affinity_diagonal_offset ust be a positive integer")

    self.spectral_min_embeddings = spectral_min_embeddings
    self.single_cluster_condition = single_cluster_condition
    self.single_cluster_affinity_threshold = single_cluster_affinity_threshold
    self.single_cluster_affinity_diagonal_offset = (
        single_cluster_affinity_diagonal_offset)
    self.fallback_clusterer_type = fallback_clusterer_type
    self.agglomerative_threshold = agglomerative_threshold
    self.naive_threshold = naive_threshold
    self.naive_adaptation_threshold = naive_adaptation_threshold


class FallbackClusterer:
  """Fallback clusterer. So far we simply use AgglomerativeClustering."""

  def __init__(self, options):
    """Initilization of the fallback clusterer.

    Args:
      options: an object of FallbackOptions
    """
    self.options = options
    if options.fallback_clusterer_type == FallbackClustererType.Agglomerative:
      self.clusterer = AgglomerativeClustering(
          n_clusters=None,
          affinity="cosine",
          linkage="average",
          distance_threshold=options.agglomerative_threshold)
    elif options.fallback_clusterer_type == FallbackClustererType.Naive:
      self.clusterer = naive_clusterer.NaiveClusterer(
          threshold=options.naive_threshold,
          adaptation_threshold=options.naive_adaptation_threshold)
    else:
      ValueError("Unsupported fallback_clusterer_type")

  def predict(self, embeddings):
    return self.clusterer.fit_predict(embeddings)


def check_single_cluster(fallback_options, embeddings, affinity):
  """Check whether this is only a single cluster.

  This function is only called when min_clusters==1.

  Args:
    fallback_options: an object of FallbackOptions
    embeddings: numpy array of shape (n_samples, n_features)
    affinity: the affinity matrix of shape (n_samples, (n_samples)

  Returns:
    a boolean, where True means there is only a single cluster
  """
  if (fallback_options.single_cluster_condition ==
      SingleClusterCondition.AllAffinity):
    if (affinity.min() >
        fallback_options.single_cluster_affinity_threshold):
      return True
  elif (fallback_options.single_cluster_condition ==
        SingleClusterCondition.NeighborAffinity):
    neighbor_affinity = np.diag(affinity, k=1)
    if (neighbor_affinity.min() >
        fallback_options.single_cluster_affinity_threshold):
      return True
  elif (fallback_options.single_cluster_condition ==
        SingleClusterCondition.AffinityStd):
    if (np.std(affinity) <
        fallback_options.single_cluster_affinity_threshold):
      return True
  elif (fallback_options.single_cluster_condition ==
        SingleClusterCondition.AffinityGmmBic):
    # Compute upper triangular matrix values to exclude diagonal values.
    if (fallback_options.single_cluster_affinity_diagonal_offset >=
        affinity.shape[0] - 1):
      raise ValueError(
          "single_cluster_affinity_diagonal_offset must be significantly "
          "smaller than affinity matrix dimension")
    upper_indices = np.triu_indices(
        affinity.shape[0],
        fallback_options.single_cluster_affinity_diagonal_offset)
    affinity_values = np.expand_dims(affinity[upper_indices], 1)

    # Fit GMM and compare BIC.
    gmm1 = GaussianMixture(n_components=1)
    gmm2 = GaussianMixture(n_components=2)
    gmm1.fit(affinity_values)
    gmm2.fit(affinity_values)
    bic1 = gmm1.bic(affinity_values)
    bic2 = gmm2.bic(affinity_values)
    return bic1 < bic2
  elif (fallback_options.single_cluster_condition ==
        SingleClusterCondition.FallbackClusterer):
    temp_clusterer = FallbackClusterer(fallback_options)
    temp_labels = temp_clusterer.predict(embeddings)
    if np.unique(temp_labels).size == 1:
      return True
  else:
    raise TypeError("Unsupported single_cluster_condition")
  return False



"""A spectral clusterer class to perform clustering."""

import numpy as np
from sklearn.cluster import AgglomerativeClustering
from spectralcluster import autotune
from spectralcluster import constraint
from spectralcluster import custom_distance_kmeans
from spectralcluster import fallback_clusterer
from spectralcluster import laplacian
from spectralcluster import refinement
from spectralcluster import utils

AutoTuneProxy = autotune.AutoTuneProxy
RefinementName = refinement.RefinementName
LaplacianType = laplacian.LaplacianType
ConstraintName = constraint.ConstraintName
EigenGapType = utils.EigenGapType


class SpectralClusterer:
  """Spectral clustering class."""

  def __init__(self,
               min_clusters=None,
               max_clusters=None,
               refinement_options=None,
               autotune=None,
               fallback_options=None,
               laplacian_type=None,
               stop_eigenvalue=1e-2,
               row_wise_renorm=False,
               custom_dist="cosine",
               max_iter=300,
               constraint_options=None,
               eigengap_type=EigenGapType.Ratio,
               max_spectral_size=None,
               affinity_function=utils.compute_affinity_matrix,
               post_eigen_cluster_function=custom_distance_kmeans.run_kmeans):
    """Constructor of the clusterer.

    Args:
      min_clusters: minimal number of clusters allowed (only effective if not
        None)
      max_clusters: maximal number of clusters allowed (only effective if not
        None), can be used together with min_clusters to fix the number of
        clusters
      refinement_options: a RefinementOptions object that contains refinement
        arguments for the affinity matrix. If None, we will not refine
      autotune: an AutoTune object to automatically search p_percentile
      fallback_options: a FallbackOptions object to indicate when to run
        fallback clusterer instead of spectral clusterer
      laplacian_type: a LaplacianType. If None, we do not use a laplacian matrix
      stop_eigenvalue: when computing the number of clusters using Eigen Gap, we
        do not look at eigen values smaller than this value
      row_wise_renorm: if True, perform row-wise re-normalization on the
        spectral embeddings
      custom_dist: str or callable. custom distance measure for k-means. If a
        string, "cosine", "euclidean", "mahalanobis", or any other distance
        functions defined in scipy.spatial.distance can be used
      max_iter: the maximum number of iterations for the custom k-means
      constraint_options: a ConstraintOptions object that contains constraint
        arguments
      eigengap_type: the type of the eigengap computation
      max_spectral_size: the maximal size of input to the spectral clustering
        algorithm. If this is set, and the actual input size is larger than
        this value, then we are going to first use hierarchical clustering
        to reduce the input size to this number. This can significantly reduce
        the computational cost for steps like Laplacian matrix and eigen
        decomposition. However, please note that this may degrade the quality
        of the final clustering results
      affinity_function: a function to compute the affinity matrix from the
        embeddings. This defaults to (cos(x,y)+1)/2
      post_eigen_cluster_function: a function to cluster the spectral embeddings
        after the eigenvalue computations. This function must have the same
        signature as custom_distance_kmeans.run_kmeans
    """
    self.min_clusters = min_clusters
    self.max_clusters = max_clusters
    if not refinement_options:
      self.refinement_options = refinement.RefinementOptions()
    else:
      self.refinement_options = refinement_options
    self.autotune = autotune
    if not fallback_options:
      self.fallback_options = fallback_clusterer.FallbackOptions()
    else:
      self.fallback_options = fallback_options
    self.laplacian_type = laplacian_type
    self.row_wise_renorm = row_wise_renorm
    self.stop_eigenvalue = stop_eigenvalue
    self.custom_dist = custom_dist
    self.max_iter = max_iter
    self.constraint_options = constraint_options
    self.eigengap_type = eigengap_type
    self.max_spectral_size = max_spectral_size
    self.affinity_function = affinity_function
    self.post_eigen_cluster_function = post_eigen_cluster_function

  def _compute_eigenvectors_ncluster(self, affinity, constraint_matrix=None):
    """Perform eigen decomposition and estiamte the number of clusters.

    Perform affinity refinement, eigen decomposition and sort eigenvectors by
    the real part of eigenvalues. Estimate the number of clusters using EigenGap
    principle.

    Args:
      affinity: the affinity matrix of input data
      constraint_matrix: numpy array of shape (n_samples, n_samples). The
        constraint matrix with prior information

    Returns:
      eigenvectors: sorted eigenvectors. numpy array of shape
      (n_samples, n_samples)
      n_clusters: number of clusters as an integer
      max_delta_norm: normalized maximum eigen gap
    """
    # Perform refinement operations on the affinity matrix.
    for refinement_name in self.refinement_options.refinement_sequence:
      refinement_operator = self.refinement_options.get_refinement_operator(
          refinement_name)
      affinity = refinement_operator.refine(affinity)

    if (self.constraint_options and
        not self.constraint_options.apply_before_refinement and
        constraint_matrix is not None):
      # Perform the constraint operation after refinement
      affinity = self.constraint_options.constraint_operator.adjust_affinity(
          affinity, constraint_matrix)

    if not self.laplacian_type or self.laplacian_type == LaplacianType.Affinity:
      # Perform eigen decomposion.
      (eigenvalues, eigenvectors) = utils.compute_sorted_eigenvectors(affinity)
      # Get number of clusters.
      n_clusters, max_delta_norm = utils.compute_number_of_clusters(
          eigenvalues,
          max_clusters=self.max_clusters,
          stop_eigenvalue=self.stop_eigenvalue,
          eigengap_type=self.eigengap_type,
          descend=True)
    else:
      # Compute Laplacian matrix
      laplacian_norm = laplacian.compute_laplacian(
          affinity, laplacian_type=self.laplacian_type)
      # Perform eigen decomposion. Eigen values are sorted in an ascending
      # order
      (eigenvalues, eigenvectors) = utils.compute_sorted_eigenvectors(
          laplacian_norm, descend=False)
      # Get number of clusters. Eigen values are sorted in an ascending order
      n_clusters, max_delta_norm = utils.compute_number_of_clusters(
          eigenvalues,
          max_clusters=self.max_clusters,
          eigengap_type=self.eigengap_type,
          descend=False)
    return eigenvectors, n_clusters, max_delta_norm

  def _reduce_size_and_predict(self, embeddings):
    """Reduce the input size, then run spectral clustering.

    Args:
      embeddings: numpy array of shape (n_samples, n_features)

    Returns:
      labels: numpy array of shape (n_samples,)
    """
    # Run AHC on the input to reduce the size.
    # Note that linkage needs to be "complete", ecause "average" and "single"
    # do not work very well here.
    # Alternatively, we can use "euclidean" and "ward", but that requires
    # that the inputs are L2 normalized first.
    ahc = AgglomerativeClustering(
        n_clusters=self.max_spectral_size,
        affinity="cosine",
        linkage="complete")
    ahc_labels = ahc.fit_predict(embeddings)

    # Compute the centroids of the AHC clusters.
    ahc_centroids = []
    for i in range(self.max_spectral_size):
      ahc_cluster_embeddings = embeddings[ahc_labels == i, :]
      ahc_centroids.append(np.mean(ahc_cluster_embeddings, axis=0))
    ahc_centroids = np.stack(ahc_centroids)

    # Run spectral clustering on AHC centroids.
    spectral_labels = self.predict(ahc_centroids)

    # Convert spectral labels to final labels.
    final_labels = np.zeros(ahc_labels.shape)
    for i in range(self.max_spectral_size):
      final_labels[ahc_labels == i] = spectral_labels[i]

    return final_labels

  def predict(self, embeddings, constraint_matrix=None):
    """Perform spectral clustering on data embeddings.

    The spectral clustering is performed on an affinity matrix.

    Args:
      embeddings: numpy array of shape (n_samples, n_features)
      constraint_matrix: numpy array of shape (n_samples, n_samples). The
        constraint matrix with prior information

    Returns:
      labels: numpy array of shape (n_samples,)

    Raises:
      TypeError: if embeddings has wrong type
      ValueError: if embeddings has wrong shape
      RuntimeError: if max_spectral_size is set and constraint_matrix is given
    """
    num_embeddings = embeddings.shape[0]

    if not isinstance(embeddings, np.ndarray):
      raise TypeError("embeddings must be a numpy array")
    if len(embeddings.shape) != 2:
      raise ValueError("embeddings must be 2-dimensional")

    # Check whether we need to run fallback clusterer instead.
    if (num_embeddings <
        self.fallback_options.spectral_min_embeddings):
      temp_clusterer = fallback_clusterer.FallbackClusterer(
          self.fallback_options)
      return temp_clusterer.predict(embeddings)

    # Check whether the input size is too big for running spectral clustering.
    if (self.max_spectral_size is not None
        and num_embeddings > self.max_spectral_size):
      if constraint_matrix is not None:
        raise RuntimeError(
            "Cannot handle constraint_matrix when max_spectral_size is set")
      if (self.max_spectral_size < 2 or
         (self.max_clusters and self.max_spectral_size <= self.max_clusters) or
         (self.min_clusters and self.max_spectral_size <= self.min_clusters)):
        raise ValueError(
            "max_spectral_size should be a relatively big number")
      return self._reduce_size_and_predict(embeddings)

    # Compute affinity matrix.
    affinity = self.affinity_function(embeddings)

    # Make single-vs-multi cluster(s) decision.
    if self.min_clusters == 1:
      if fallback_clusterer.check_single_cluster(
              self.fallback_options, embeddings, affinity):
        return np.array([0] * num_embeddings)

    # Apply constraint.
    if (self.constraint_options and
        self.constraint_options.apply_before_refinement and
        constraint_matrix is not None):
      # Perform the constraint operation before refinement
      affinity = self.constraint_options.constraint_operator.adjust_affinity(
          affinity, constraint_matrix)

    if self.autotune:
      # Use Auto-tuning method to find a good p_percentile.
      if (RefinementName.RowWiseThreshold
          not in self.refinement_options.refinement_sequence):
        raise ValueError(
            "AutoTune is only effective when the refinement sequence"
            "contains RowWiseThreshold")

      def p_percentile_to_ratio(p_percentile):
        """Compute the `ratio` given a `p_percentile` value."""
        self.refinement_options.p_percentile = p_percentile
        (eigenvectors, n_clusters,
         max_delta_norm) = self._compute_eigenvectors_ncluster(
             affinity, constraint_matrix)
        if self.autotune.proxy == AutoTuneProxy.PercentileSqrtOverNME:
          ratio = np.sqrt(1 - p_percentile) / max_delta_norm
        elif self.autotune.proxy == AutoTuneProxy.PercentileSqrtOverNME:
          ratio = (1 - p_percentile) / max_delta_norm
        else:
          raise ValueError("Unsupported value of AutoTuneProxy")
        return ratio, eigenvectors, n_clusters

      eigenvectors, n_clusters, _ = self.autotune.tune(p_percentile_to_ratio)
    else:
      # Do not use Auto-tune.
      eigenvectors, n_clusters, _ = self._compute_eigenvectors_ncluster(
          affinity, constraint_matrix)

    if self.min_clusters is not None:
      n_clusters = max(n_clusters, self.min_clusters)

    # Get spectral embeddings.
    spectral_embeddings = eigenvectors[:, :n_clusters]

    if self.row_wise_renorm:
      # Perform row wise re-normalization.
      rows_norm = np.linalg.norm(spectral_embeddings, axis=1, ord=2)
      spectral_embeddings = spectral_embeddings / np.reshape(
          rows_norm, (num_embeddings, 1))

    # Run clustering algorithm on spectral embeddings. This defaults
    # to customized K-means.
    labels = self.post_eigen_cluster_function(
        spectral_embeddings=spectral_embeddings,
        n_clusters=n_clusters,
        custom_dist=self.custom_dist,
        max_iter=self.max_iter)
    return labels
